class Client:
    def get(self, url, data=None):
        """
        Opens url using GET.

        Return a Page() instance.
        """

        raise NotImplementedError

    def post(self, url, data=None):
        """
        Opens url using POST.

        Return a Page() instance.
        """

        raise NotImplementedError


from sulfur.errors import ValidationError


def check_response(url, url_data=None,
                   method='GET', post=None,
                   login=None, login_required=False,
                   codes=None, code_checker=None,
                   html5=False, html5_validator=None,
                   xhtml=False,
                   client=None):
    """
    Assert that response obey certain constraints.

    Args:
        url (str):
            Page's url. It uses url_data to format the url as
            ``url.format(**url_data)``. If a list of urls is given, it tests each
            url separately.
        url_data (dict):
            A dictionary used to format url strings.
        method (str):
            HTTP method used to retrieve page. Must be a string 'GET' (default)
            or 'POST'.
        post (dict):
            A dictionary with post data. If given, makes method='POST'.
        codes (sequence):
            a list of allowed HTTP response codes.
        code_checker (function):
            a function that checks if the response code is valid or not.
        login (str or tuple):
            Can be either a username or a tuple with (username, password) used
            to login just before making the request.
        login_required (bool):
            If True, anonymous users must either return a 401, 403, 404 or it
            must be a redirect to a login page.
        html5 (bool):
            If True, makes sure that the response is valid HTML5.
        html5_validator:
            Method used to validate the HTML5 code.
        xhtml (bool):
            If True, makes sure the response is valid XHTML.

    Raises:
        If check fails, raises an AssertionError.
    """

    # If a list of urls is given, call check_response() in each element
    # separately
    if not isinstance(url, (str, bytes)):
        kwargs = locals()
        del kwargs['url']
        for url_item in url:
            print('checking', url_item)
            check_response(url_item, **kwargs)
        return

    # Format url using url_data
    if url_data:
        url = url.format(**url_data)

    # Makes login, if necessary
    client = _get_client_object(client)
    if login:
        if isinstance(login, (tuple, list)):
            username, password = login
            client.login(username=username, password=password)
        else:
            client.force_login(login)

    # Build kwargs for executing client's .get(), .post() or other HTTP methods
    client_kwargs = {}
    if method == 'POST' and post:
        client_kwargs['data'] = post

    # Fetch data from server object
    http_method = getattr(client, method.lower())
    response = http_method(url, **client_kwargs)
    response_data = response.content

    # Normalize code checker to be a function that returns True on valid
    # response codes. Check if status code is valid
    if codes and not code_checker:
        def code_checker(code):
            return code in codes
    if not code_checker:
        def code_checker(code):
            return True
    status_code = response.status_code
    if not code_checker(status_code):
        raise ValidationError('%s: received invalid status code: %s' %
                              (url, status_code))

    # Now we check if the content HTML data validates
    if html5:
        html5_validator = _get_html5_validator(html5_validator)
        try:
            html5_validator(response_data)
        except ValidationError as ex:
            raise ValidationError('%s: %s' % (url, ex))


def _get_client_object(server, **kwargs):
    """
    Return the server object from request.
    """

    if server is not None:
        return server

    # Global server object
    from django.test.client import Client

    return Client(**kwargs)


def _get_html5_validator(html5_validator):
    """
    Normalize the html5_validator parameter from check_response.
    """

    from sulfur.validators import Html5Validator

    if html5_validator is None:
        html5_validator = 'default'
    if callable(html5_validator):
        return html5_validator
    return Html5Validator.as_validator(html5_validator)


def check_ok(url, **kwargs):
    """
    Like :func:`check_response`, but checks if response code is either in the
    2xx or in the 3xx range.
    """

    @_set_func_kwargs(kwargs)
    def code_checker(code):
        return 200 <= code <= 399

    return check_response(url, **kwargs)


def check_2xx(url, **kwargs):
    """
    Like :func:`check_response`, but checks if response code is in the 2xx
    range.
    """

    @_set_func_kwargs(kwargs)
    def code_checker(code):
        return 200 <= code <= 299

    return check_response(url, **kwargs)


def check_3xx(url, **kwargs):
    """
    Like :func:`check_response`, but checks if response code is in the 3xx
    range.
    """

    @_set_func_kwargs(kwargs)
    def code_checker(code):
        return 300 <= code <= 399

    return check_response(url, **kwargs)


def check_4xx(url, **kwargs):
    """
    Like :func:`check_response`, but checks if response code is in the 4xx
    range.
    """

    @_set_func_kwargs(kwargs)
    def code_checker(code):
        return 400 <= code <= 499

    return check_response(url, **kwargs)


def check_404(url, **kwargs):
    """
    Like :func:`check_response`, but checks if response code is 404.
    """

    return check_response(url, codes=[404], **kwargs)


def check_5xx(url, **kwargs):
    """
    Like :func:`check_response`, but checks if response code is in the 5xx
    range.
    """

    @_set_func_kwargs(kwargs)
    def code_checker(code):
        return 500 <= code <= 599

    return check_response(url, **kwargs)


# Aliases
def check_success(url, **kwargs):
    """
    Alias to :func:`check_2xx`.
    """
    return check_2xx(url, **kwargs)


def check_redirect(url, **kwargs):
    """
    Alias to :func:`check_3xx`.
    """
    return check_3xx(url, **kwargs)


def check_client_error(url, **kwargs):
    """
    Alias to :func:`check_4xx`.
    """
    return check_4xx(url, **kwargs)


def check_server_error(url, **kwargs):
    """
    Alias to :func:`check_5xx`.
    """
    return check_5xx(url, **kwargs)


def _set_func_kwargs(dic):
    """
    Usage::

        @_set_func_kwargs(D)
        def func():
            pass

    Same as D['func'] = func
    """

    def decorator(func):
        dic[func.__name__] = func
        return func

    return decorator



import argparse
import sulfur


# Create an argument parser.
parser = argparse.ArgumentParser('sulfur')


def main(args=None):
    """
    Main entry point for your project.

    Parameters:
        args : list
            A of arguments as if they were input in the command line. Leave it
            None use sys.argv.
    """

    args = parser.parse_args(args)

    print('Not implemented yet! ;)')
    raise SystemExit


if __name__ == '__main__':
    main()


from collections import Sequence

from bs4 import BeautifulSoup
from lazyutils import lazy
import selenium.webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC


class Query(Sequence):
    def __init__(self, parent, elements):
        self.parent = parent
        self.elements = elements

    def click(self):
        """Clicks on all selected elements."""
        [x.click() for x in self]
        return self

    def __iter__(self):
        return iter(self.elements)

    def __len__(self):
        return len(self.elements)

    def __getitem__(self, i):
        return self.elements[i]


class IdManager:
    def __init__(self, driver):
        self.__dict__['_IdManager__driver'] = driver

    def __getattr__(self, id):
        return self[id]

    def __setattr__(self, id, value):
        self[id] = value

    def __getitem__(self, id):
        return self.__driver.find_element_by_id(id)

    def __setitem__(self, id, value):
        self[id].send_keys(value)


class WebDriverFacade:
    """
    A simplified interface to the selenium webdriver with simpler method names
    and a few useful resources.
    """

    soup_lib = 'html5lib'

    def __init__(self, driver, base_url='', wait=0, **kwargs):
        self._driver = driver
        self.base_url = base_url
        self.implicitly_wait(wait)
        for k, v in kwargs:
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise TypeError('invalid argument: %r' % k)

    def __getattr__(self, attr):
        return getattr(self._driver, attr)

    def __call__(self, selector):
        elements = self.find_elements_by_css_selector(selector)
        return Query(self, [ElementFacade(x) for x in elements])

    def __getitem__(self, selector):
        return ElementFacade(self.find_element_by_css_selector(selector))

    @property
    def selenium(self):
        return self._driver

    @lazy
    def soup(self):
        """
        A beautiful soup interface to the HTML source code
        """

        return BeautifulSoup(self.html, self.soup_lib)

    @property
    def id(self):
        """
        A simplified interface to dom elements with a defined id.
        """

        return IdManager(self)

    def get(self, url=''):
        self.url = self.base_url + url
        return self._driver.get(self.base_url + url)

    def click(self, selector):
        """Clicks in the first element with the given CSS selector."""

        self[selector].click()

    def keys(self, *args):
        """Alias to send_keys()"""

        return self.send_keys(*args)

    # Wait conditions
    def wait_on(self, func, timeout=1.0):
        """Wait until func(driver) return True.

        Raises a TimeoutError if condition is not met in the given interval."""

        WebDriverWait(self, timeout).until(func)

    def wait_title_is(self, value, timeout=1.0):
        """Waits until the page title assumes the given value.

        Raises a TimeoutError if condition is not met in the given interval."""

        condition = EC.title_is(value)
        WebDriverWait(self, timeout).until(condition)

    def wait_title_contains(self, value, timeout=1.0):
        """Waits until the page title contains the given value.

        Raises a TimeoutError if condition is not met in the given interval."""

        condition = EC.title_contains(value)
        WebDriverWait(self, timeout).until(condition)


class ElementFacade:
    def __init__(self, element):
        self._element = element

    def fill(self, data):
        self.send_keys(data)

    def __getattr__(self, attr):
        return getattr(self._element, attr)


class Driver(WebDriverFacade):
    """
    The sulfur web driver.

    Args:
        driver:
            If given, can be a string or a Selenium webdriver object. The
            accept strings are: "firefox", "chromium".
    """
    def __init__(self, driver=None, **kwargs):
        if driver is None:
            driver = selenium.webdriver.Firefox()
        if isinstance(driver, str):
            driver_cls = get_driver_class_from_string()
            driver = driver_cls()
        super().__init__(driver, **kwargs)


def get_driver_class_from_string(name):
    """
    Select driver class from name.
    """

    mapping = {
        'firefox': 'selenium.webdriver.Firefox',
        'chrome': 'selenium.webdriver.Chrome',
        'ie': 'selenium.webdriver.Ie',
        'edge': 'selenium.webdriver.Edge',
        'opera': 'selenium.webdriver.Opera',
        'safari': 'selenium.webdriver.Safari',
        'blackberry': 'selenium.webdriver.BlackBerry',
        'phantomjs': 'selenium.webdriver.PhantomJS',
        'android': 'selenium.webdriver.Android',
    }

    mod, _, cls = mapping[name].rpartition('.')
    cls_list = __import__(mod, fromlist=[cls])
    return cls_list[0]


from html5lib import HTMLParser
from html5lib.html5parser import ParseError

from sulfur.errors import ValidationError
from sulfur.simplehtml5 import SimpleHtml5Validator


class Html5Validator:
    """
    Validates HTML5 data using a series of methods.
    """

    @classmethod
    def as_validator(cls, name):
        """
        Return a validator function from the given validator method name.
        """

        try:
            getattr(cls, name + '_validator')
        except AttributeError:
            raise ValueError('invalid validator: %r' % name)

        def validator(data):
            instance = cls(data)
            method = getattr(instance, name + '_validator')
            return method()

        return validator

    def __init__(self, data):
        self.data = data

    def default_validator(self):
        """
        Default validation method.

        Just check if tags are well formed and if all tags that should close,
        indeed close.
        """

        validator = SimpleHtml5Validator(self.data)
        validator.validate()

    def strict_validator(self):
        """
        Strict validation method.

        We just call html5lib parser with strict=True. Error messages are awful,
        and it complaints about many small errors, so it can be annoying.
        """

        strict_parser = HTMLParser(strict=True)
        try:
            strict_parser.parse(self.data)
        except ParseError as ex:
            raise ValidationError(str(ex))





import collections
import re

from sulfur.errors import ValidationError

# Generic regex
spaces_re = re.compile(r'\s+')

# Opening tag processors
tag_separator_re = re.compile(r'(?:\s*/>|\s*>|\s+)')
opt_spaces_re = re.compile(r'\s*')
tagname_re = re.compile(r'\w([-a-zA-Z]*\w+)?')

# Attribute names
attr_name_re = re.compile(r'(\w|_)([-a-zA-Z_]*(\w|_)+)?')
eq_sign_re = re.compile(r'\s*=\s*')
string_re = re.compile(r'"[^"]*"')
attr_value_re = re.compile(r'(\w|_)([-a-zA-Z_]*(\w|_)+)?')

# CLosing tags
self_closing_tags = {'br', 'link', 'meta'}
CLOSE_TAG = object()


class ValidMixin:
    """
    Implement the as_valid() and validate() methods.
    """

    @classmethod
    def as_valid(cls, *args, **kwargs):
        """
        Create a new instance and immediately validate it.

        Arguments are passed to the constructor.
        """

        instance = cls(*args, **kwargs)
        instance.validate()
        return instance

    def validate(self):
        """
        Validate element.

        Raises an ValidationError() if validation fails.
        """

        raise ValidationError('validation must be implemented in subclasses')


class HtmlElement(ValidMixin):
    """
    Base HTML class for Tags/Document and other HTML elements.
    """

    def __init__(self, children=None):
        self.children = list(children or [])

    def add(self, tag):
        """
        Add a new child element to itself.
        """

        if isinstance(tag, Text) and tag.isspace() or tag is None:
            return
        self.children.append(tag)

    def render(self):
        """
        Renders HTML structure as string.
        """

        return ''.join(c.render() for c in self.children)


class Document(HtmlElement):
    """
    Root node for the HTML AST.
    """

    def __repr__(self):
        return '<html>'


class HtmlTag(HtmlElement):
    """
    Represents a generic HTML tag.
    """

    @property
    def self_closing(self):
        """
        If True, means that the tag is self-closing (e.g. <br>).
        """

        return self.tag in self_closing_tags

    def __init__(self, tag, children=None, attributes=None):
        super().__init__(children)
        self.tag = tag
        self.attributes = list(attributes or [])

    def __repr__(self):
        return '<%s>' % self.tag

    def validate(self):
        validate_tag_name(self.tag)

    def render_attributes(self):
        """
        Render the attributes part of string.
        """

        if not self.attributes:
            return ''
        else:
            data = []
            for name, value in self.attributes:
                if value is None:
                    data.append(' %s' % name)
                else:
                    data.append(' %s=%s' % (name, value))
            return ''.join(data)

    def render(self):
        if self.self_closing and not self.children:
            return '<%s>' % self.tag
        children = super().render()
        attr_data = self.render_attributes()
        return '<%s%s>%s</%s>' % (self.tag, attr_data, children, self.tag)


class TextBase(ValidMixin, collections.UserString):
    """
    Base class for HTML text fragments.
    """

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.data)

    def validate(self):
        if '<' in self.data or '>' in self.data:
            raise ValidationError('forgot to escape <> inside text')

    def render(self):
        return str(self)


class DOCTYPE(TextBase):
    """
    The doctype declaration on top of the HTML source file.

    HTML5 doctype is simply this::

        <!DOCTYPE html>
    """

    def validate(self):
        if self != 'html':
            raise ValidationError('please use the <!DOCTYPE html> declaration '
                                  'for html5 documents.')

    def render(self):
        return '<!DOCTYPE %s>' % self.data


class Comment(TextBase):
    """
    Data inside HTML comments::

        <!-- This is an HTML comment -->
    """

    def validate(self):
        if '-->' in self:
            raise ValidationError('invalid comment: closing twice with "-->"')

    def render(self):
        return '<!--%s-->' % self


class Text(TextBase):
    """
    A text fragment in an HTML source.
    """

    def validate(self):
        if tagname_re.fullmatch(self.tag) is None:
            raise ValidationError('invalid tag name: %r' % self.tag)


class SimpleHtml5Validator:
    """
    Base HTML5 validator class.

    Simple API::

        validator = SimpleHtml5Validator(source)
        validator.validate()

    Users should use the validate() function. This class exists publicly to be
    subclassed.
    """

    def __init__(self, data, doctype=False, head=False, body=False):
        self.data = data
        self.doctype = doctype
        self.head = head
        self.body = body
        self._stream = self.data
        self._document = Document()
        self._open_tags = [self._document]
        self._current = self._document

    def _parse_comments(self):
        ...
        return

    def _parse_doctype(self):
        if not self._stream.startswith('<!DOCTYPE'):
            return None

        data, sep, self._stream = self._stream.partition('>')
        if not sep:
            raise ValidationError('invalid doctype declaration: ' + data)
        data = data[9:]  # remove <!DOCTYPE string from be￼ginning
        return DOCTYPE.as_valid(data.strip())

    def _parse_spaces(self):
        data = self._parse_from_re(spaces_re)
        if data:
            return Text.as_valid(data)
        return None

    def _parse_text(self):
        data, sep, missing = self._stream.partition('<')
        self._stream = sep + missing
        return Text(data)

    def _parse_tag(self):
        tagname, sep, self._stream = re_partition(tag_separator_re,
                                                  self._stream)
        tagname += sep.strip()
        tagname = tagname[1:]
        tag_kwargs = {}
        if tagname.endswith('/>'):
            tagname = tagname[:-2]
        elif tagname.endswith('>'):
            tagname = tagname[:-1]
        else:
            tag_kwargs['attributes'] = attributes = []
            while not (self._stream.startswith('>') or
                           self._stream.startswith('/>')):

                # We start reading an attribute name and then process it
                # looking for a possible value in the rhs of the equality
                name, self._stream = read_tok(attr_name_re, self._stream)
                if not name:
                    raise ValidationError(
                        'expect attribute on open "%s" tag' % tagname)

                # Now we look for the "=" sign to decide if we must check the
                # rhs of the attribute declaration
                op, self._stream = read_tok(eq_sign_re, self._stream)
                if not op:
                    value = None
                else:
                    if self._stream.startswith('"'):
                        value, self._stream = read_tok(string_re, self._stream)
                    else:
                        old = self._stream
                        value, sep, self._stream = re_partition(tag_separator_re, self._stream)
                        if sep.strip():
                            self._stream = sep.strip() + self._stream

                attributes.append((name, value))

                if not self._stream:
                    raise ValidationError(
                        '">" expected for closing %s tag' % tagname)
            self._stream = self._stream[1:]  # strip the first ">" character

        tag = HtmlTag.as_valid(tagname.strip(), **tag_kwargs)
        self._open_tags.append(tag)
        return tag

    def _parse_endtag(self):
        tag, _, self._stream = self._stream.partition('>')
        tagname = tag[2:]  # remove "</"
        if not tagname_re.fullmatch(tagname):
            raise ValidationError('invalid closing tag name: %s' % tagname)

        while self._open_tags:
            last = self._open_tags.pop()
            if last.tag == tagname:
                break
            elif last.self_closing:
                continue
            raise ValidationError('forgot to close tag' % last.tag)
        else:
            raise ValidationError(
                'closing tag </%s> without an open tag' % tagname
            )
        return CLOSE_TAG

    def _parse_from_re(self, regex):
        match = regex.match(self._stream)
        if match is None:
            return False

        end = match.end()
        data = self._stream[:end]
        self._stream = self._stream[end:]
        return data

    def _parse_next(self):
        data = self._stream
        if data.startswith('<!--'):
            return self._parse_comments()
        if data.startswith('<!'):
            return self._parse_doctype()
        elif data.startswith('</'):
            return self._parse_endtag()
        elif data.startswith('<'):
            return self._parse_tag()
        else:
            return self._parse_text()

    def validate(self):
        """
        Parse document and raise a ValidationError in case of an invalid
        document.
        """

        while self._stream:
            elem = self._parse_next()
            if elem is CLOSE_TAG:
                self._current = self._open_tags[-1]
            else:
                self._current.add(elem)
                if isinstance(elem, HtmlTag):
                    self._current = elem

        return self._document


def re_partition(re, st):
    match = re.search(st)
    if match is None:
        return st, '', ''
    i = match.start()
    j = match.end()
    if i == j == 0:
        return st, '', ''
    return st[:i], st[i:j], st[j:]


def read_tok(re, st):
    match = re.match(st)
    if match:
        return st[:match.end()], st[match.end():]
    else:
        return '', st


def validate_tag_name(tagname):
    """
    Raise ValidationError() if tagname is not valid.
    """

    if tagname_re.fullmatch(tagname) is None:
        raise ValidationError('invalid tag name: %r' % tagname)


def validate(html_str, **kwargs):
    """
    Validate string of HTML source.

    Args:
        html_str (str):
            A string of HTML source.
        doctype:
            If True, ensures that the document has the HTML5 doctype.
        head:
            If True, ensures that document has a <head> element. (This is *not*
            required by the HTML5 spec, but most users want to enforce that).
        body:
            If True, ensures that document has a <body> element. (This is *not*
            required by the HTML5 spec, but most users want to enforce that).

    Returns:
        An Document() instance with the document AST.
    """

    validator = SimpleHtml5Validator(html_str, **kwargs)
    return validator.validate()



import pytest
from sulfur.simplehtml5 import validate


def assert_ok(data, html_identity=True):
    ast = validate(data)
    if html_identity:
        html_render = ast.render().strip()
        data = data.strip()
        assert html_render == data


def test_validate_simple_doctype_document():
    assert_ok("<!DOCTYPE html>")


def test_validate_simple_document():
    assert_ok("<!DOCTYPE html><head></head><body></body>")


def test_validate_tag_with_attribute():
    assert_ok('<div id="foo"></div>')


def test_validate_tag_with_unquoted_attribute():
    assert_ok('<div id=foo></div>')


def test_validate_tag_with_boolean_attribute():
    assert_ok('<div is-a-tag></div>')


def test_validate_self_closing_tag():
    assert_ok('<p>foo<br></p>')
    assert_ok('<p>foo<br/></p>', False)




import pytest


@pytest.fixture
def sulfur():
    return 'sulfur'


@pytest.fixture
def urlchecker():
    """
    Return the sulfur.urlchecker module.
    """

    import sulfur.urlchecker
    return sulfur.urlchecker
